
import pytest
import responses

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from instabot.api.config import API_URL, SIG_KEY_VERSION

from .test_bot import TestBot
from .test_variables import (TEST_USERNAME_INFO_ITEM, TEST_PHOTO_ITEM, TEST_CAPTION_ITEM, TEST_COMMENT_ITEM,
                             TEST_SEARCH_USERNAME_ITEM, TEST_FOLLOWER_ITEM, TEST_FOLLOWING_ITEM, TEST_TIMELINE_PHOTO_ITEM)


class TestBotGet(TestBot):
    @pytest.mark.parametrize(
        'media_id,check_media,'
        'comment_txt,'
        'has_liked,'
        'like_count,'
        'has_anonymous_profile_picture,filter_users_without_profile_photo,'
        'expected', [
            (1234567890, False, False, True, float('inf'), True, True, True),
            (1234567890, False, False, True, float('inf'), True, False, True),
            (1234567890, False, False, True, float('inf'), False, True, True),
            (1234567890, False, False, True, float('inf'), False, False, True),
            (1234567890, True, False, True, float('inf'), True, True, False),
            (1234567890, True, False, True, float('inf'), True, False, False),
            (1234567890, True, False, True, float('inf'), False, True, False),
            (1234567890, True, False, True, float('inf'), False, False, False),
            (1234567890, False, False, False, float('inf'), True, True, True),
            (1234567890, False, False, False, float('inf'), True, False, True),
            (1234567890, False, False, False, float('inf'), False, True, True),
            (1234567890, False, False, False, float('inf'), False, False, True),
            (1234567890, True, False, False, float('inf'), True, True, False),
            (1234567890, True, False, False, float('inf'), True, False, False),
            (1234567890, True, False, False, float('inf'), False, True, False),
            (1234567890, True, False, False, float('inf'), False, False, False),
            (1234567890, False, False, True, False, True, True, True),
            (1234567890, False, False, True, False, True, False, True),
            (1234567890, False, False, True, False, False, True, True),
            (1234567890, False, False, True, False, False, False, True),
            (1234567890, True, False, True, False, True, True, False),
            (1234567890, True, False, True, False, True, False, False),
            (1234567890, True, False, True, False, False, True, False),
            (1234567890, True, False, True, False, False, False, False),
            (1234567890, False, False, False, False, True, True, True),
            (1234567890, False, False, False, False, True, False, True),
            (1234567890, False, False, False, False, False, True, True),
            (1234567890, False, False, False, False, False, False, True),
            (1234567890, True, False, False, False, True, True, False),
            (1234567890, True, False, False, False, True, False, True),
            (1234567890, True, False, False, False, False, True, True),
            (1234567890, True, False, False, False, False, False, True),
            (1234567890, False, True, True, float('inf'), True, True, True),
            (1234567890, False, True, True, float('inf'), True, False, True),
            (1234567890, False, True, True, float('inf'), False, True, True),
            (1234567890, False, True, True, float('inf'), False, False, True),
            (1234567890, True, True, True, float('inf'), True, True, False),
            (1234567890, True, True, True, float('inf'), True, False, False),
            (1234567890, True, True, True, float('inf'), False, True, False),
            (1234567890, True, True, True, float('inf'), False, False, False),
            (1234567890, False, True, False, float('inf'), True, True, True),
            (1234567890, False, True, False, float('inf'), True, False, True),
            (1234567890, False, True, False, float('inf'), False, True, True),
            (1234567890, False, True, False, float('inf'), False, False, True),
            (1234567890, True, True, False, float('inf'), True, True, False),
            (1234567890, True, True, False, float('inf'), True, False, False),
            (1234567890, True, True, False, float('inf'), False, True, False),
            (1234567890, True, True, False, float('inf'), False, False, False),
            (1234567890, False, True, True, False, True, True, True),
            (1234567890, False, True, True, False, True, False, True),
            (1234567890, False, True, True, False, False, True, True),
            (1234567890, False, True, True, False, False, False, True),
            (1234567890, True, True, True, False, True, True, False),
            (1234567890, True, True, True, False, True, False, False),
            (1234567890, True, True, True, False, False, True, False),
            (1234567890, True, True, True, False, False, False, False),
            (1234567890, False, True, False, False, True, True, True),
            (1234567890, False, True, False, False, True, False, True),
            (1234567890, False, True, False, False, False, True, True),
            (1234567890, False, True, False, False, False, False, True),
            (1234567890, True, True, False, False, True, True, False),
            (1234567890, True, True, False, False, True, False, False),
            (1234567890, True, True, False, False, False, True, False),
            (1234567890, True, True, False, False, False, False, False)])
    @responses.activate
    @patch('time.sleep', return_value=None)
    def test_bot_like(self, patched_time_sleep, media_id, check_media,
                      comment_txt,
                      has_liked,
                      like_count,
                      has_anonymous_profile_picture, filter_users_without_profile_photo,
                      expected):

        self.bot._following = [1]
        TEST_PHOTO_ITEM['has_liked'] = has_liked
        if not like_count:
            like_count = self.bot.min_likes_to_like + 1
        TEST_PHOTO_ITEM['like_count'] = like_count
        TEST_PHOTO_ITEM['user']['pk'] = self.bot.user_id + 1
        TEST_USERNAME_INFO_ITEM['pk'] = self.bot.user_id + 2
        TEST_USERNAME_INFO_ITEM['follower_count'] = 100
        TEST_USERNAME_INFO_ITEM['following_count'] = 15
        TEST_USERNAME_INFO_ITEM['has_anonymous_profile_picture'] = has_anonymous_profile_picture
        self.bot.filter_users_without_profile_photo = filter_users_without_profile_photo
        TEST_USERNAME_INFO_ITEM['is_business'] = False
        TEST_USERNAME_INFO_ITEM['is_private'] = False
        TEST_USERNAME_INFO_ITEM['is_verified'] = False
        TEST_USERNAME_INFO_ITEM['media_count'] = self.bot.min_media_count_to_follow + 1
        if comment_txt:
            comment_txt = ' '.join(self.bot.blacklist_hashtags)
        else:
            comment_txt = 'instabot'
        TEST_USERNAME_INFO_ITEM['biography'] = comment_txt

        responses.add(
            responses.GET, "{api_url}media/{media_id}/info/".format(
                api_url=API_URL, media_id=media_id),
            json={
                "auto_load_more_enabled": True,
                "num_results": 1,
                "status": "ok",
                "more_available": False,
                "items": [TEST_PHOTO_ITEM]
            }, status=200)

        responses.add(
            responses.GET, "{api_url}media/{media_id}/info/".format(
                api_url=API_URL, media_id=media_id),
            json={
                "auto_load_more_enabled": True,
                "num_results": 1,
                "status": "ok",
                "more_available": False,
                "items": [TEST_PHOTO_ITEM]
            }, status=200)

        results = 1
        response_data = {
            "caption": TEST_CAPTION_ITEM,
            "caption_is_edited": False,
            "comment_count": 4,
            "comment_likes_enabled": True,
            "comments": [TEST_COMMENT_ITEM for _ in range(results)],
            "has_more_comments": False,
            "has_more_headload_comments": False,
            "media_header_display": "none",
            "preview_comments": [],
            "status": "ok"
        }
        responses.add(
            responses.GET, '{api_url}media/{media_id}/comments/?'.format(
                api_url=API_URL, media_id=media_id), json=response_data, status=200)

        responses.add(
            responses.GET, "{api_url}media/{media_id}/info/".format(
                api_url=API_URL, media_id=media_id),
            json={
                "auto_load_more_enabled": True,
                "num_results": 1,
                "status": "ok",
                "more_available": False,
                "items": [TEST_PHOTO_ITEM]
            }, status=200)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=TEST_PHOTO_ITEM['user']['pk']
            ), status=200, json=response_data)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=TEST_PHOTO_ITEM['user']['pk']
            ), status=200, json=response_data)

        responses.add(
            responses.POST, '{api_url}media/{media_id}/like/'.format(
                api_url=API_URL, media_id=media_id
            ), status=200, json={'status': 'ok'})

        assert self.bot.like(media_id, check_media=check_media) == expected

    @pytest.mark.parametrize(
        'comment_id', [12345678901234567, '12345678901234567'])
    @responses.activate
    def test_bot_like_comment(self, comment_id):
        responses.add(
            responses.POST, '{api_url}media/{comment_id}/comment_like/'.format(
                api_url=API_URL, comment_id=comment_id
            ), json={'status': 'ok'}, status=200
        )
        assert self.bot.like_comment(comment_id)

    @responses.activate
    @pytest.mark.parametrize('has_liked_comment,comment_id', [
        (True, True),
        (True, False),
        (False, False),
        (False, True)])
    @patch('time.sleep', return_value=None)
    def test_like_media_comments(self, patched_time_sleep, has_liked_comment, comment_id):
        TEST_COMMENT_ITEM['has_liked_comment'] = has_liked_comment
        results = 2
        if comment_id or has_liked_comment:
            comment_id = TEST_COMMENT_ITEM['pk']
            expected_broken_items = []
        else:
            comment_id = 'wrong_comment_id'
            expected_broken_items = [TEST_COMMENT_ITEM['pk'] for _ in range(results)]
        response_data = {
            "caption": TEST_CAPTION_ITEM,
            "caption_is_edited": False,
            "comment_count": 4,
            "comment_likes_enabled": True,
            "comments": [TEST_COMMENT_ITEM for _ in range(results)],
            "has_more_comments": False,
            "has_more_headload_comments": False,
            "media_header_display": "none",
            "preview_comments": [],
            "status": "ok"
        }
        media_id = 1234567890
        responses.add(
            responses.GET, '{api_url}media/{media_id}/comments/?'.format(
                api_url=API_URL, media_id=media_id), json=response_data, status=200)
        responses.add(
            responses.POST, '{api_url}media/{comment_id}/comment_like/'.format(
                api_url=API_URL, comment_id=comment_id
            ), json={'status': 'ok'}, status=200
        )
        broken_items = self.bot.like_media_comments(media_id)
        assert broken_items == expected_broken_items

    @responses.activate
    @pytest.mark.parametrize('user_id', ['1234567890', 1234567890])
    @patch('time.sleep', return_value=None)
    def test_like_user(self, patched_time_sleep, user_id):
        self.bot._following = [1]

        TEST_USERNAME_INFO_ITEM['biography'] = 'instabot'

        response_data = {
            'status': 'ok',
            'user': TEST_SEARCH_USERNAME_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{username}/usernameinfo/'.format(
                api_url=API_URL, username=user_id
            ), status=200, json=response_data)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=user_id
            ), status=200, json=response_data)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=user_id
            ), status=200, json=response_data)

        results = 5
        response_data = {
            "auto_load_more_enabled": True,
            "num_results": results,
            "status": "ok",
            "more_available": False,
            "items": [TEST_PHOTO_ITEM for _ in range(results)]
        }
        responses.add(
            responses.GET, '{api_url}feed/user/{user_id}/?max_id={max_id}&min_timestamp={min_timestamp}&rank_token={rank_token}&ranked_content=true'.format(
                api_url=API_URL, user_id=user_id, max_id='',
                min_timestamp=None, rank_token=self.bot.api.rank_token),
            json=response_data, status=200)

        responses.add(
            responses.GET, "{api_url}media/{media_id}/info/".format(
                api_url=API_URL, media_id=TEST_PHOTO_ITEM['pk']),
            json={
                "auto_load_more_enabled": True,
                "num_results": 1,
                "status": "ok",
                "more_available": False,
                "items": [TEST_PHOTO_ITEM]
            }, status=200)

        results = 2
        response_data = {
            "caption": TEST_CAPTION_ITEM,
            "caption_is_edited": False,
            "comment_count": 4,
            "comment_likes_enabled": True,
            "comments": [TEST_COMMENT_ITEM for _ in range(results)],
            "has_more_comments": False,
            "has_more_headload_comments": False,
            "media_header_display": "none",
            "preview_comments": [],
            "status": "ok"
        }
        responses.add(
            responses.GET, '{api_url}media/{media_id}/comments/?'.format(
                api_url=API_URL, media_id=TEST_PHOTO_ITEM['pk']), json=response_data, status=200)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=TEST_PHOTO_ITEM['user']['pk']
            ), status=200, json=response_data)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=TEST_PHOTO_ITEM['user']['pk']
            ), status=200, json=response_data)

        responses.add(
            responses.POST, '{api_url}media/{media_id}/like/'.format(
                api_url=API_URL, media_id=TEST_PHOTO_ITEM['pk']
            ), status=200, json={'status': 'ok'})

        broken_items = self.bot.like_user(user_id)
        assert [] == broken_items

    @responses.activate
    @pytest.mark.parametrize('user_ids', [['1234567890'], [1234567890]])
    @patch('time.sleep', return_value=None)
    def test_like_users(self, patched_time_sleep, user_ids):

        self.bot._following = [1]

        TEST_USERNAME_INFO_ITEM['biography'] = 'instabot'

        response_data = {
            'status': 'ok',
            'user': TEST_SEARCH_USERNAME_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{username}/usernameinfo/'.format(
                api_url=API_URL, username=user_ids[0]
            ), status=200, json=response_data)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=user_ids[0]
            ), status=200, json=response_data)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=user_ids[0]
            ), status=200, json=response_data)

        results_1 = 5
        response_data = {
            "auto_load_more_enabled": True,
            "num_results": results_1,
            "status": "ok",
            "more_available": False,
            "items": [TEST_PHOTO_ITEM for _ in range(results_1)]
        }
        responses.add(
            responses.GET, '{api_url}feed/user/{user_id}/?max_id={max_id}&min_timestamp={min_timestamp}&rank_token={rank_token}&ranked_content=true'.format(
                api_url=API_URL, user_id=user_ids[0], max_id='',
                min_timestamp=None, rank_token=self.bot.api.rank_token),
            json=response_data, status=200)

        responses.add(
            responses.GET, "{api_url}media/{media_id}/info/".format(
                api_url=API_URL, media_id=TEST_PHOTO_ITEM['pk']),
            json={
                "auto_load_more_enabled": True,
                "num_results": 1,
                "status": "ok",
                "more_available": False,
                "items": [TEST_PHOTO_ITEM]
            }, status=200)

        results_2 = 2
        response_data = {
            "caption": TEST_CAPTION_ITEM,
            "caption_is_edited": False,
            "comment_count": results_2,
            "comment_likes_enabled": True,
            "comments": [TEST_COMMENT_ITEM for _ in range(results_2)],
            "has_more_comments": False,
            "has_more_headload_comments": False,
            "media_header_display": "none",
            "preview_comments": [],
            "status": "ok"
        }
        responses.add(
            responses.GET, '{api_url}media/{media_id}/comments/?'.format(
                api_url=API_URL, media_id=TEST_PHOTO_ITEM['pk']), json=response_data, status=200)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=TEST_PHOTO_ITEM['user']['pk']
            ), status=200, json=response_data)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=TEST_PHOTO_ITEM['user']['pk']
            ), status=200, json=response_data)

        responses.add(
            responses.POST, '{api_url}media/{media_id}/like/'.format(
                api_url=API_URL, media_id=TEST_PHOTO_ITEM['pk']
            ), status=200, json={'status': 'ok'})

        self.bot.like_users(user_ids)
        assert self.bot.total['likes'] == results_1

    @responses.activate
    @pytest.mark.parametrize('medias', [[1234567890, 9876543210]])
    @patch('time.sleep', return_value=None)
    def test_like_medias(self, patched_time_sleep, medias):
        self.bot._following = [1]

        for media in medias:
            TEST_PHOTO_ITEM['pk'] = media
            responses.add(
                responses.GET, "{api_url}media/{media_id}/info/".format(
                    api_url=API_URL, media_id=media),
                json={
                    "auto_load_more_enabled": True,
                    "num_results": 1,
                    "status": "ok",
                    "more_available": False,
                    "items": [TEST_PHOTO_ITEM]
                }, status=200)

            results = 2
            response_data = {
                "caption": TEST_CAPTION_ITEM,
                "caption_is_edited": False,
                "comment_count": 4,
                "comment_likes_enabled": True,
                "comments": [TEST_COMMENT_ITEM for _ in range(results)],
                "has_more_comments": False,
                "has_more_headload_comments": False,
                "media_header_display": "none",
                "preview_comments": [],
                "status": "ok"
            }
            responses.add(
                responses.GET, '{api_url}media/{media_id}/comments/?'.format(
                    api_url=API_URL, media_id=TEST_PHOTO_ITEM['pk']), json=response_data, status=200)

            response_data = {
                'status': 'ok',
                'user': TEST_USERNAME_INFO_ITEM
            }
            responses.add(
                responses.GET, '{api_url}users/{user_id}/info/'.format(
                    api_url=API_URL, user_id=TEST_PHOTO_ITEM['user']['pk']
                ), status=200, json=response_data)

            response_data = {
                'status': 'ok',
                'user': TEST_USERNAME_INFO_ITEM
            }
            responses.add(
                responses.GET, '{api_url}users/{user_id}/info/'.format(
                    api_url=API_URL, user_id=TEST_PHOTO_ITEM['user']['pk']
                ), status=200, json=response_data)

            responses.add(
                responses.POST, '{api_url}media/{media_id}/like/'.format(
                    api_url=API_URL, media_id=TEST_PHOTO_ITEM['pk']
                ), status=200, json={'status': 'ok'})

        broken_items = self.bot.like_medias(medias)
        assert [] == broken_items

    @responses.activate
    @pytest.mark.parametrize('hashtag', ['like_hashtag1', 'like_hashtag2'])
    @patch('time.sleep', return_value=None)
    def test_like_hashtag(self, patche_time_sleep, hashtag):
        self.bot._following = [1]
        liked_at_start = self.bot.total['likes']
        results_1 = 10
        my_test_photo_item = TEST_PHOTO_ITEM.copy()
        my_test_photo_item["like_count"] = self.bot.min_likes_to_like + 1
        my_test_photo_item["has_liked"] = False
        response_data = {
            "auto_load_more_enabled": True,
            "num_results": results_1,
            "status": "ok",
            "more_available": True,
            "next_max_id": my_test_photo_item['id'],
            "items": [my_test_photo_item for _ in range(results_1)]
        }
        responses.add(
            responses.GET, '{api_url}feed/tag/{hashtag}/?max_id={max_id}&rank_token={rank_token}&ranked_content=true&'.format(
                api_url=API_URL, hashtag=hashtag, max_id='',
                rank_token=self.bot.api.rank_token),
            json=response_data, status=200)

        responses.add(
            responses.GET, "{api_url}media/{media_id}/info/".format(
                api_url=API_URL, media_id=my_test_photo_item['pk']),
            json={
                "auto_load_more_enabled": True,
                "num_results": 1,
                "status": "ok",
                "more_available": False,
                "items": [my_test_photo_item]
            }, status=200)

        results_2 = 2
        response_data = {
            "caption": TEST_CAPTION_ITEM,
            "caption_is_edited": False,
            "comment_count": results_2,
            "comment_likes_enabled": True,
            "comments": [TEST_COMMENT_ITEM for _ in range(results_2)],
            "has_more_comments": False,
            "has_more_headload_comments": False,
            "media_header_display": "none",
            "preview_comments": [],
            "status": "ok"
        }
        responses.add(
            responses.GET, '{api_url}media/{media_id}/comments/?'.format(
                api_url=API_URL, media_id=TEST_PHOTO_ITEM['pk']), json=response_data, status=200)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=my_test_photo_item['user']['pk']
            ), status=200, json=response_data)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=my_test_photo_item['user']['pk']
            ), status=200, json=response_data)

        responses.add(
            responses.POST, '{api_url}media/{media_id}/like/'.format(
                api_url=API_URL, media_id=my_test_photo_item['pk']
            ), status=200, json={'status': 'ok'})

        broken_items = self.bot.like_hashtag(hashtag)
        assert [] == broken_items
        assert self.bot.total['likes'] == liked_at_start + results_1

    @responses.activate
    @pytest.mark.parametrize('username', [
        '1234567890', 1234567890
    ])
    @patch('time.sleep', return_value=None)
    def test_like_followers(self, patched_time_sleep, username):

        liked_at_start = self.bot.total['likes']

        test_username = 'test.username'

        response_data_1 = {
            'status': 'ok',
            'user': TEST_SEARCH_USERNAME_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{username}/usernameinfo/'.format(
                api_url=API_URL, username=test_username
            ), status=200, json=response_data_1)

        response_data_2 = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=username
            ), status=200, json=response_data_2)

        results_3 = 2
        response_data_3 = {
            'status': 'ok',
            'big_list': False,
            'next_max_id': None,
            'sections': None,
            'users': [TEST_FOLLOWER_ITEM for _ in range(results_3)]
        }
        responses.add(
            responses.GET, "{api_url}friendships/{user_id}/followers/?rank_token={rank_token}".format(
                api_url=API_URL, user_id=username, rank_token=self.bot.api.rank_token
            ), json=response_data_3, status=200)

        self.bot._following = [1]

        TEST_USERNAME_INFO_ITEM['biography'] = 'instabot'
        my_test_photo_item = TEST_PHOTO_ITEM.copy()
        my_test_photo_item["like_count"] = self.bot.min_likes_to_like + 1
        my_test_photo_item["has_liked"] = False

        response_data = {
            'status': 'ok',
            'user': TEST_SEARCH_USERNAME_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{username}/usernameinfo/'.format(
                api_url=API_URL, username=username
            ), status=200, json=response_data)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=username
            ), status=200, json=response_data)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=username
            ), status=200, json=response_data)

        results_4 = 3
        response_data = {
            "auto_load_more_enabled": True,
            "num_results": results_4,
            "status": "ok",
            "more_available": False,
            "items": [my_test_photo_item for _ in range(results_4)]
        }
        responses.add(
            responses.GET, '{api_url}feed/user/{user_id}/?max_id={max_id}&min_timestamp={min_timestamp}&rank_token={rank_token}&ranked_content=true'.format(
                api_url=API_URL, user_id=username, max_id='',
                min_timestamp=None, rank_token=self.bot.api.rank_token),
            json=response_data, status=200)

        responses.add(
            responses.GET, "{api_url}media/{media_id}/info/".format(
                api_url=API_URL, media_id=my_test_photo_item['pk']),
            json={
                "auto_load_more_enabled": True,
                "num_results": 1,
                "status": "ok",
                "more_available": False,
                "items": [my_test_photo_item]
            }, status=200)

        results_5 = 2
        response_data = {
            "caption": TEST_CAPTION_ITEM,
            "caption_is_edited": False,
            "comment_count": results_5,
            "comment_likes_enabled": True,
            "comments": [TEST_COMMENT_ITEM for _ in range(results_5)],
            "has_more_comments": False,
            "has_more_headload_comments": False,
            "media_header_display": "none",
            "preview_comments": [],
            "status": "ok"
        }
        responses.add(
            responses.GET, '{api_url}media/{media_id}/comments/?'.format(
                api_url=API_URL, media_id=my_test_photo_item['pk']), json=response_data, status=200)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=my_test_photo_item['user']['pk']
            ), status=200, json=response_data)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=my_test_photo_item['user']['pk']
            ), status=200, json=response_data)

        responses.add(
            responses.POST, '{api_url}media/{media_id}/like/'.format(
                api_url=API_URL, media_id=my_test_photo_item['pk']
            ), status=200, json={'status': 'ok'})

        self.bot.like_followers(username)
        assert self.bot.total['likes'] == liked_at_start + results_3 * results_4

    @responses.activate
    @pytest.mark.parametrize('username', [
        '1234567890', 1234567890
    ])
    @patch('time.sleep', return_value=None)
    def test_like_following(self, patched_time_sleep, username):

        liked_at_start = self.bot.total['likes']

        test_username = 'test.username'

        response_data_1 = {
            'status': 'ok',
            'user': TEST_SEARCH_USERNAME_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{username}/usernameinfo/'.format(
                api_url=API_URL, username=test_username
            ), status=200, json=response_data_1)

        response_data_2 = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=username
            ), status=200, json=response_data_2)

        results_3 = 5
        response_data_3 = {
            'status': 'ok',
            'big_list': False,
            'next_max_id': None,
            'sections': None,
            'users': [TEST_FOLLOWING_ITEM for _ in range(results_3)]
        }
        responses.add(
            responses.GET, "{api_url}friendships/{user_id}/following/?max_id={max_id}&ig_sig_key_version={sig_key}&rank_token={rank_token}".format(
                api_url=API_URL, user_id=username, rank_token=self.bot.api.rank_token, sig_key=SIG_KEY_VERSION, max_id=''
            ), json=response_data_3, status=200)

        self.bot._following = [1]

        TEST_USERNAME_INFO_ITEM['biography'] = 'instabot'
        my_test_photo_item = TEST_PHOTO_ITEM.copy()
        my_test_photo_item["like_count"] = self.bot.min_likes_to_like + 1
        my_test_photo_item["has_liked"] = False

        response_data = {
            'status': 'ok',
            'user': TEST_SEARCH_USERNAME_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{username}/usernameinfo/'.format(
                api_url=API_URL, username=username
            ), status=200, json=response_data)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=username
            ), status=200, json=response_data)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=username
            ), status=200, json=response_data)

        results_4 = 3
        response_data = {
            "auto_load_more_enabled": True,
            "num_results": results_4,
            "status": "ok",
            "more_available": False,
            "items": [my_test_photo_item for _ in range(results_4)]
        }
        responses.add(
            responses.GET, '{api_url}feed/user/{user_id}/?max_id={max_id}&min_timestamp={min_timestamp}&rank_token={rank_token}&ranked_content=true'.format(
                api_url=API_URL, user_id=username, max_id='',
                min_timestamp=None, rank_token=self.bot.api.rank_token),
            json=response_data, status=200)

        responses.add(
            responses.GET, "{api_url}media/{media_id}/info/".format(
                api_url=API_URL, media_id=my_test_photo_item['pk']),
            json={
                "auto_load_more_enabled": True,
                "num_results": 1,
                "status": "ok",
                "more_available": False,
                "items": [my_test_photo_item]
            }, status=200)

        results_5 = 2
        response_data = {
            "caption": TEST_CAPTION_ITEM,
            "caption_is_edited": False,
            "comment_count": results_5,
            "comment_likes_enabled": True,
            "comments": [TEST_COMMENT_ITEM for _ in range(results_5)],
            "has_more_comments": False,
            "has_more_headload_comments": False,
            "media_header_display": "none",
            "preview_comments": [],
            "status": "ok"
        }
        responses.add(
            responses.GET, '{api_url}media/{media_id}/comments/?'.format(
                api_url=API_URL, media_id=my_test_photo_item['pk']), json=response_data, status=200)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=my_test_photo_item['user']['pk']
            ), status=200, json=response_data)

        response_data = {
            'status': 'ok',
            'user': TEST_USERNAME_INFO_ITEM
        }
        responses.add(
            responses.GET, '{api_url}users/{user_id}/info/'.format(
                api_url=API_URL, user_id=my_test_photo_item['user']['pk']
            ), status=200, json=response_data)

        responses.add(
            responses.POST, '{api_url}media/{media_id}/like/'.format(
                api_url=API_URL, media_id=my_test_photo_item['pk']
            ), status=200, json={'status': 'ok'})

        self.bot.like_following(username)
        assert self.bot.total['likes'] == liked_at_start + results_3 * results_4

    @responses.activate
    @patch('time.sleep', return_value=None)
    def test_like_timeline(self, patched_time_sleep):

        my_test_timelime_photo_item = TEST_TIMELINE_PHOTO_ITEM.copy()
        my_test_timelime_photo_item['media_or_ad']["like_count"] = self.bot.max_likes_to_like - 1
        my_test_timelime_photo_item['media_or_ad']["has_liked"] = False

        liked_at_start = self.bot.total['likes']

        results_1 = 8
        responses.add(
            responses.GET, "{api_url}feed/timeline/".format(api_url=API_URL),
            json={
                "auto_load_more_enabled": True,
                "num_results": results_1,
                "is_direct_v2_enabled": True,
                "status": "ok",
                "next_max_id": None,
                "more_available": False,
                "feed_items": [my_test_timelime_photo_item for _ in range(results_1)]
            }, status=200)

        responses.add(
            responses.POST, '{api_url}media/{media_id}/like/'.format(
                api_url=API_URL, media_id=my_test_timelime_photo_item['media_or_ad']['pk']
            ), status=200, json={'status': 'ok'})

        broken_items = self.bot.like_timeline()
        assert [] == broken_items
        assert self.bot.total['likes'] == liked_at_start + results_1



import datetime
import os


def get_tsv_line(dictionary):
    line = ""
    for key in sorted(dictionary):
        line += str(dictionary[key]) + "\t"
    return line[:-1] + "\n"


def get_header_line(dictionary):
    line = "\t".join(sorted(dictionary))
    return line + "\n"


def ensure_dir(file_path):
    directory = os.path.dirname(file_path)
    if not os.path.exists(directory) and directory:
        os.makedirs(directory)


def dump_data(data, path):
    ensure_dir(path)
    if not os.path.exists(path):
        with open(path, "w") as f:
            f.write(get_header_line(data))
            f.write(get_tsv_line(data))
    else:
        with open(path, "a") as f:
            f.write(get_tsv_line(data))


def save_user_stats(self, username, path=""):
    if not username:
        username = self.api.username
    user_id = self.convert_to_user_id(username)
    infodict = self.get_user_info(user_id, use_cache=False)
    if infodict:
        data_to_save = {
            "date": str(datetime.datetime.now().replace(microsecond=0)),
            "followers": int(infodict["follower_count"]),
            "following": int(infodict["following_count"]),
            "medias": int(infodict["media_count"])
        }
        file_path = os.path.join(path, "%s.tsv" % username)
        dump_data(data_to_save, file_path)
        self.logger.info("Stats saved at %s." % data_to_save["date"])
    return False



import atexit
import datetime
import random
import signal
import time

from .. import utils
from ..api import API
from .bot_archive import archive, archive_medias, unarchive_medias
from .bot_block import block, block_bots, block_users, unblock, unblock_users
from .bot_checkpoint import load_checkpoint, save_checkpoint
from .bot_comment import (comment, comment_geotag, comment_hashtag,
                          comment_medias, comment_user, comment_users,
                          is_commented, reply_to_comment)
from .bot_delete import delete_comment, delete_media, delete_medias
from .bot_direct import (send_hashtag, send_like, send_media, send_medias,
                         send_message, send_messages, send_profile)
from .bot_filter import check_media, check_not_bot, check_user, filter_medias
from .bot_follow import (follow, follow_followers, follow_following,
                         follow_users)
from .bot_get import (convert_to_user_id, get_archived_medias, get_comment,
                      get_comment_likers, get_geotag_medias, get_geotag_users,
                      get_hashtag_medias, get_hashtag_users,
                      get_last_user_medias, get_locations_from_coordinates,
                      get_media_commenters, get_media_comments,
                      get_media_comments_all, get_media_id_from_link,
                      get_link_from_media_id, get_media_info, get_media_likers,
                      get_media_owner, get_messages, get_popular_medias,
                      get_timeline_medias, get_timeline_users,
                      get_total_hashtag_medias, get_total_user_medias,
                      get_user_followers, get_user_following,
                      get_user_id_from_username, get_user_info,
                      get_user_likers, get_user_medias, get_user_tags_medias,
                      get_username_from_user_id, get_your_medias, search_users)
from .bot_like import (like, like_comment, like_followers, like_following,
                       like_geotag, like_hashtag, like_media_comments,
                       like_medias, like_timeline, like_user, like_users)
from .bot_photo import download_photo, download_photos, upload_photo
from .bot_stats import save_user_stats
from .bot_support import (check_if_file_exists, console_print, extract_urls,
                          read_list_from_file)
from .bot_unfollow import (unfollow, unfollow_everyone, unfollow_non_followers,
                           unfollow_users)
from .bot_unlike import (unlike, unlike_comment, unlike_media_comments,
                         unlike_medias, unlike_user)
from .bot_video import upload_video


class Bot(object):
    def __init__(self,
                 whitelist_file='whitelist.txt',
                 blacklist_file='blacklist.txt',
                 comments_file='comments.txt',
                 followed_file='followed.txt',
                 unfollowed_file='unfollowed.txt',
                 skipped_file='skipped.txt',
                 friends_file='friends.txt',
                 proxy=None,
                 max_likes_per_day=1000,
                 max_unlikes_per_day=1000,
                 max_follows_per_day=350,
                 max_unfollows_per_day=350,
                 max_comments_per_day=100,
                 max_blocks_per_day=100,
                 max_unblocks_per_day=100,
                 max_likes_to_like=100,
                 min_likes_to_like=20,
                 max_messages_per_day=300,
                 filter_users=True,
                 filter_private_users=True,
                 filter_users_without_profile_photo=True,
                 filter_previously_followed=False,
                 filter_business_accounts=True,
                 filter_verified_accounts=True,
                 max_followers_to_follow=2000,
                 min_followers_to_follow=10,
                 max_following_to_follow=2000,
                 min_following_to_follow=10,
                 max_followers_to_following_ratio=10,
                 max_following_to_followers_ratio=2,
                 min_media_count_to_follow=3,
                 max_following_to_block=2000,
                 like_delay=10,
                 unlike_delay=10,
                 follow_delay=30,
                 unfollow_delay=30,
                 comment_delay=60,
                 block_delay=30,
                 unblock_delay=30,
                 message_delay=60,
                 stop_words=('shop', 'store', 'free'),
                 blacklist_hashtags=['#shop', '#store', '#free'],
                 verbosity=True,
                 device=None
                 ):
        self.api = API(device=device)

        self.total = {'likes': 0,
                      'unlikes': 0,
                      'follows': 0,
                      'unfollows': 0,
                      'comments': 0,
                      'blocks': 0,
                      'unblocks': 0,
                      'messages': 0,
                      'archived': 0,
                      'unarchived': 0}

        self.start_time = datetime.datetime.now()

        self.delays = {'like': like_delay,
                       'unlike': unlike_delay,
                       'follow': follow_delay,
                       'unfollow': unfollow_delay,
                       'comment': comment_delay,
                       'block': block_delay,
                       'unblock': unblock_delay,
                       'message': message_delay}

        self.last = {key: 0 for key in self.delays.keys()}

        # limits - follow
        self.filter_users = filter_users
        self.filter_private_users = filter_private_users
        self.filter_users_without_profile_photo = filter_users_without_profile_photo
        self.filter_business_accounts = filter_business_accounts
        self.filter_verified_accounts = filter_verified_accounts
        self.filter_previously_followed = filter_previously_followed

        self.max_per_day = {'likes': max_likes_per_day,
                            'unlikes': max_unlikes_per_day,
                            'follows': max_follows_per_day,
                            'unfollows': max_unfollows_per_day,
                            'comments': max_comments_per_day,
                            'blocks': max_blocks_per_day,
                            'unblocks': max_unblocks_per_day,
                            'messages': max_messages_per_day}

        self.max_likes_to_like = max_likes_to_like
        self.min_likes_to_like = min_likes_to_like
        self.max_followers_to_follow = max_followers_to_follow
        self.min_followers_to_follow = min_followers_to_follow
        self.max_following_to_follow = max_following_to_follow
        self.min_following_to_follow = min_following_to_follow
        self.max_followers_to_following_ratio = max_followers_to_following_ratio
        self.max_following_to_followers_ratio = max_following_to_followers_ratio
        self.min_media_count_to_follow = min_media_count_to_follow
        self.stop_words = stop_words
        self.blacklist_hashtags = blacklist_hashtags

        # limits - block
        self.max_following_to_block = max_following_to_block

        # current following and followers
        self._following = None
        self._followers = None
        self._user_infos = {}  # User info cache
        self._usernames = {}  # `username` to `user_id` mapping

        # Database files
        self.followed_file = utils.file(followed_file)
        self.unfollowed_file = utils.file(unfollowed_file)
        self.skipped_file = utils.file(skipped_file)
        self.friends_file = utils.file(friends_file)
        self.comments_file = utils.file(comments_file)
        self.blacklist_file = utils.file(blacklist_file)
        self.whitelist_file = utils.file(whitelist_file)

        self.proxy = proxy
        self.verbosity = verbosity

        self.logger = self.api.logger
        self.logger.info('Instabot Started')

    @property
    def user_id(self):
        # For compatibility
        return self.api.user_id

    @property
    def username(self):
        # For compatibility
        return self.api.username

    @property
    def password(self):
        # For compatibility
        return self.api.password

    @property
    def last_json(self):
        # For compatibility
        return self.api.last_json

    @property
    def blacklist(self):
        # This is a fast operation because `get_user_id_from_username` is cached.
        return [self.convert_to_user_id(i) for i in self.blacklist_file.list
                if i is not None]

    @property
    def whitelist(self):
        # This is a fast operation because `get_user_id_from_username` is cached.
        return [self.convert_to_user_id(i) for i in self.whitelist_file.list
                if i is not None]

    @property
    def following(self):
        now = time.time()
        last = self.last.get('updated_following', now)
        if self._following is None or now - last > 7200:
            self.console_print('`bot.following` is empty, will download.', 'green')
            self._following = self.get_user_following(self.user_id)
            self.last['updated_following'] = now
        return self._following

    @property
    def followers(self):
        now = time.time()
        last = self.last.get('updated_followers', now)
        if self._followers is None or now - last > 7200:
            self.console_print('`bot.followers` is empty, will download.', 'green')
            self._followers = self.get_user_followers(self.user_id)
            self.last['updated_followers'] = now
        return self._followers

    def version(self):
        try:
            from pip._vendor import pkg_resources
        except ImportError:
            import pkg_resources
        return next((p.version for p in pkg_resources.working_set if p.project_name.lower() == 'instabot'), "No match")

    def logout(self, *args, **kwargs):
        save_checkpoint(self)
        self.api.logout()
        self.logger.info("Bot stopped. "
                         "Worked: %s", datetime.datetime.now() - self.start_time)
        self.print_counters()

    def login(self, **args):
        if self.proxy:
            args['proxy'] = self.proxy
        if self.api.login(**args) is False:
            return False
        self.prepare()
        signal.signal(signal.SIGTERM, self.logout)
        atexit.register(self.logout)
        return True

    def prepare(self):
        storage = load_checkpoint(self)
        if storage is not None:
            self.total, self.api.total_requests, self.start_time = storage

    def print_counters(self):
        for key, val in self.total.items():
            if val > 0:
                self.logger.info("Total {}: {}{}".format(key, val,
                                                         "/" + str(self.max_per_day[key]) if self.max_per_day.get(key) else ""))
        self.logger.info("Total requests: {}".format(self.api.total_requests))

    def delay(self, key):
        """Sleep only if elapsed time since `self.last[key]` < `self.delay[key]`."""
        last_action, target_delay = self.last[key], self.delays[key]
        elapsed_time = time.time() - last_action
        if elapsed_time < target_delay:
            t_remaining = target_delay - elapsed_time
            time.sleep(t_remaining * random.uniform(0.25, 1.25))
        self.last[key] = time.time()

    def error_delay(self):
        time.sleep(10)

    def small_delay(self):
        time.sleep(random.uniform(0.75, 3.75))

    def very_small_delay(self):
        time.sleep(random.uniform(0.175, 0.875))

    def reached_limit(self, key):
        current_date = datetime.datetime.now()
        passed_days = (current_date.date() - self.start_time.date()).days
        if passed_days > 0:
            self.reset_counters()
        return self.max_per_day[key] - self.total[key] <= 0

    def reset_counters(self):
        for k in self.total:
            self.total[k] = 0
        self.start_time = datetime.datetime.now()

    # getters

    def get_your_medias(self, as_dict=False):
        """
        Returns your media ids. With parameter as_dict=True returns media as dict.
        :type as_dict: bool
        """
        return get_your_medias(self, as_dict)

    def get_archived_medias(self, as_dict=False):
        """
        Returns your archived media ids. With parameter as_dict=True returns media as dict.
        :type as_dict: bool
        """
        return get_archived_medias(self, as_dict)

    def get_timeline_medias(self):
        return get_timeline_medias(self)

    def get_popular_medias(self):
        return get_popular_medias(self)

    def get_user_medias(self, user_id, filtration=True, is_comment=False):
        return get_user_medias(self, user_id, filtration, is_comment)

    def get_total_user_medias(self, user_id):
        return get_total_user_medias(self, user_id)

    def get_last_user_medias(self, user_id, count):
        """
        Returns the last number of posts specified in count in media ids array.
        :type count: int
        :param count: Count of posts
        :return: array
        """
        return get_last_user_medias(self, user_id, count)

    def get_hashtag_medias(self, hashtag, filtration=True):
        return get_hashtag_medias(self, hashtag, filtration)

    def get_total_hashtag_medias(self, hashtag, amount=100, filtration=False):
        return get_total_hashtag_medias(self, hashtag, amount, filtration)

    def get_geotag_medias(self, geotag, filtration=True):
        return get_geotag_medias(self, geotag, filtration)

    def get_locations_from_coordinates(self, latitude, longitude):
        return get_locations_from_coordinates(self, latitude, longitude)

    def get_media_info(self, media_id):
        return get_media_info(self, media_id)

    def get_timeline_users(self):
        return get_timeline_users(self)

    def get_hashtag_users(self, hashtag):
        return get_hashtag_users(self, hashtag)

    def get_geotag_users(self, geotag):
        return get_geotag_users(self, geotag)

    def get_user_id_from_username(self, username):
        return get_user_id_from_username(self, username)

    def get_user_tags_medias(self, user_id):
        return get_user_tags_medias(self, user_id)

    def get_username_from_user_id(self, user_id):
        return get_username_from_user_id(self, user_id)

    def get_user_info(self, user_id, use_cache=True):
        return get_user_info(self, user_id, use_cache)

    def get_user_followers(self, user_id, nfollows=None):
        return get_user_followers(self, user_id, nfollows)

    def get_user_following(self, user_id, nfollows=None):
        return get_user_following(self, user_id, nfollows)

    def get_comment_likers(self, comment_id):
        return get_comment_likers(self, comment_id)

    def get_media_likers(self, media_id):
        return get_media_likers(self, media_id)

    def get_media_comments(self, media_id, only_text=False):
        return get_media_comments(self, media_id, only_text)

    def get_media_comments_all(self, media_id, only_text=False, count=False):
        return get_media_comments_all(self, media_id, only_text, count)

    def get_comment(self):
        return get_comment(self)

    def get_media_commenters(self, media_id):
        return get_media_commenters(self, media_id)

    def get_media_owner(self, media):
        return get_media_owner(self, media)

    def get_user_likers(self, user_id, media_count=10):
        return get_user_likers(self, user_id, media_count)

    def get_media_id_from_link(self, link):
        return get_media_id_from_link(self, link)

    def get_link_from_media_id(self, link):
        return get_link_from_media_id(self, link)

    def get_messages(self):
        return get_messages(self)

    def search_users(self, query):
        return search_users(self, query)

    def convert_to_user_id(self, usernames):
        return convert_to_user_id(self, usernames)

    # like

    def like(self, media_id, check_media=True):
        return like(self, media_id, check_media)

    def like_comment(self, comment_id):
        return like_comment(self, comment_id)

    def like_medias(self, media_ids, check_media=True):
        return like_medias(self, media_ids, check_media)

    def like_timeline(self, amount=None):
        return like_timeline(self, amount)

    def like_media_comments(self, media_id):
        return like_media_comments(self, media_id)

    def like_user(self, user_id, amount=None, filtration=True):
        return like_user(self, user_id, amount, filtration)

    def like_hashtag(self, hashtag, amount=None):
        return like_hashtag(self, hashtag, amount)

    def like_geotag(self, geotag, amount=None):
        return like_geotag(self, geotag, amount)

    def like_users(self, user_ids, nlikes=None, filtration=True):
        return like_users(self, user_ids, nlikes, filtration)

    def like_followers(self, user_id, nlikes=None, nfollows=None):
        return like_followers(self, user_id, nlikes, nfollows)

    def like_following(self, user_id, nlikes=None, nfollows=None):
        return like_following(self, user_id, nlikes, nfollows)

    # unlike

    def unlike(self, media_id):
        return unlike(self, media_id)

    def unlike_comment(self, comment_id):
        return unlike_comment(self, comment_id)

    def unlike_media_comments(self, media_id):
        return unlike_media_comments(self, media_id)

    def unlike_medias(self, media_ids):
        return unlike_medias(self, media_ids)

    def unlike_user(self, user):
        return unlike_user(self, user)

    # photo

    def download_photo(self, media_id, folder='photos', filename=None, save_description=False):
        return download_photo(self, media_id, folder, filename, save_description)

    def download_photos(self, medias, folder='photos', save_description=False):
        return download_photos(self, medias, folder, save_description)

    def upload_photo(self, photo, caption=None, upload_id=None, from_video=False):
        return upload_photo(self, photo, caption, upload_id, from_video)

    # video

    def upload_video(self, video, caption=''):
        return upload_video(self, video, caption)

    # follow

    def follow(self, user_id):
        return follow(self, user_id)

    def follow_users(self, user_ids):
        return follow_users(self, user_ids)

    def follow_followers(self, user_id, nfollows=None):
        return follow_followers(self, user_id, nfollows)

    def follow_following(self, user_id, nfollows=None):
        return follow_following(self, user_id, nfollows)

    # unfollow

    def unfollow(self, user_id):
        return unfollow(self, user_id)

    def unfollow_users(self, user_ids):
        return unfollow_users(self, user_ids)

    def unfollow_non_followers(self, n_to_unfollows=None):
        return unfollow_non_followers(self, n_to_unfollows)

    def unfollow_everyone(self):
        return unfollow_everyone(self)

    # direct

    def send_message(self, text, user_ids, thread_id=None):
        return send_message(self, text, user_ids, thread_id)

    def send_messages(self, text, user_ids):
        return send_messages(self, text, user_ids)

    def send_media(self, media_id, user_ids, text=None, thread_id=None):
        return send_media(self, media_id, user_ids, text, thread_id)

    def send_medias(self, media_id, user_ids, text=None):
        return send_medias(self, media_id, user_ids, text)

    def send_hashtag(self, hashtag, user_ids, text='', thread_id=None):
        return send_hashtag(self, hashtag, user_ids, text, thread_id)

    def send_profile(self, profile_user_id, user_ids, text='', thread_id=None):
        return send_profile(self, profile_user_id, user_ids, text, thread_id)

    def send_like(self, user_ids, thread_id=None):
        return send_like(self, user_ids, thread_id)

    # delete

    def delete_media(self, media_id):
        return delete_media(self, media_id)

    def delete_medias(self, medias):
        return delete_medias(self, medias)

    def delete_comment(self, media_id, comment_id):
        return delete_comment(self, media_id, comment_id)

    # archive

    def archive(self, media_id, undo=False):
        return archive(self, media_id, undo)

    def unarchive(self, media_id):
        return archive(self, media_id, True)

    def archive_medias(self, medias):
        return archive_medias(self, medias)

    def unarchive_medias(self, medias):
        return unarchive_medias(self, medias)

    # comment

    def comment(self, media_id, comment_text):
        return comment(self, media_id, comment_text)

    def reply_to_comment(self, media_id, comment_text, parent_comment_id):
        return reply_to_comment(self, media_id, comment_text, parent_comment_id)

    def comment_hashtag(self, hashtag, amount=None):
        return comment_hashtag(self, hashtag, amount)

    def comment_medias(self, medias):
        return comment_medias(self, medias)

    def comment_user(self, user_id, amount=None):
        return comment_user(self, user_id, amount)

    def comment_users(self, user_ids, ncomments=None):
        return comment_users(self, user_ids, ncomments)

    def comment_geotag(self, geotag):
        return comment_geotag(self, geotag)

    def is_commented(self, media_id):
        return is_commented(self, media_id)

    # block

    def block(self, user_id):
        return block(self, user_id)

    def unblock(self, user_id):
        return unblock(self, user_id)

    def block_users(self, user_ids):
        return block_users(self, user_ids)

    def unblock_users(self, user_ids):
        return unblock_users(self, user_ids)

    def block_bots(self):
        return block_bots(self)

    # filter

    def filter_medias(self, media_items, filtration=True, quiet=False, is_comment=False):
        return filter_medias(self, media_items, filtration, quiet, is_comment)

    def check_media(self, media):
        return check_media(self, media)

    def check_user(self, user, unfollowing=False):
        return check_user(self, user, unfollowing)

    def check_not_bot(self, user):
        return check_not_bot(self, user)

    # support

    def check_if_file_exists(self, file_path, quiet=False):
        return check_if_file_exists(file_path, quiet)

    def extract_urls(self, text):
        return extract_urls(text)

    def read_list_from_file(self, file_path):
        return read_list_from_file(file_path)

    def console_print(self, text, color=None):
        return console_print(self, text, color)

    # stats

    def save_user_stats(self, username, path=""):
        return save_user_stats(self, username, path=path)



import time

from tqdm import tqdm


def follow(self, user_id):
    user_id = self.convert_to_user_id(user_id)
    msg = ' ===> Going to follow `user_id`: {}.'.format(user_id)
    self.console_print(msg)
    if not self.check_user(user_id):
        return False
    if not self.reached_limit('follows'):
        self.delay('follow')
        if self.api.follow(user_id):
            msg = '===> FOLLOWED <==== `user_id`: {}.'.format(user_id)
            self.console_print(msg, 'green')
            self.total['follows'] += 1
            self.followed_file.append(user_id)
            if user_id not in self.following:
                self.following.append(user_id)
            return True
    else:
        self.logger.info("Out of follows for today.")
    return False


def follow_users(self, user_ids):
    broken_items = []
    if self.reached_limit('follows'):
        self.logger.info("Out of follows for today.")
        return
    msg = "Going to follow {} users.".format(len(user_ids))
    self.logger.info(msg)
    skipped = self.skipped_file
    followed = self.followed_file
    unfollowed = self.unfollowed_file
    self.console_print(msg, 'green')

    # Remove skipped and already followed and unfollowed list from user_ids
    user_ids = list(set(user_ids) - skipped.set - followed.set - unfollowed.set)
    msg = 'After filtering followed, unfollowed and `{}`, {} user_ids left to follow.'
    msg = msg.format(skipped.fname, len(user_ids))
    self.console_print(msg, 'green')
    for user_id in tqdm(user_ids, desc='Processed users'):
        if self.reached_limit('follows'):
            self.logger.info("Out of follows for today.")
            break
        if not self.follow(user_id):
            if self.api.last_response.status_code == 404:
                self.console_print("404 error user {user_id} doesn't exist.", 'red')
                broken_items.append(user_id)

            elif self.api.last_response.status_code == 200:
                broken_items.append(user_id)

            elif self.api.last_response.status_code not in (400, 429):
                # 400 (block to follow) and 429 (many request error)
                # which is like the 500 error.
                try_number = 3
                error_pass = False
                for _ in range(try_number):
                    time.sleep(60)
                    error_pass = self.follow(user_id)
                    if error_pass:
                        break
                if not error_pass:
                    self.error_delay()
                    i = user_ids.index(user_id)
                    broken_items += user_ids[i:]
                    break

    self.logger.info("DONE: Now following {} users in total.".format(self.total['follows']))
    return broken_items


def follow_followers(self, user_id, nfollows=None):
    self.logger.info("Follow followers of: {}".format(user_id))
    if self.reached_limit('follows'):
        self.logger.info("Out of follows for today.")
        return
    if not user_id:
        self.logger.info("User not found.")
        return
    followers = self.get_user_followers(user_id, nfollows)
    followers = list(set(followers) - set(self.blacklist))
    if not followers:
        self.logger.info("{} not found / closed / has no followers.".format(user_id))
    else:
        self.follow_users(followers[:nfollows])


def follow_following(self, user_id, nfollows=None):
    self.logger.info("Follow following of: {}".format(user_id))
    if self.reached_limit('follows'):
        self.logger.info("Out of follows for today.")
        return
    if not user_id:
        self.logger.info("User not found.")
        return
    followings = self.get_user_following(user_id)
    if not followings:
        self.logger.info("{} not found / closed / has no following.".format(user_id))
    else:
        self.follow_users(followings[:nfollows])



from tqdm import tqdm


def archive(self, media_id, undo=False):
    self.small_delay()
    media = self.get_media_info(media_id)
    media = media[0] if isinstance(media, list) else media
    if self.api.archive_media(media, undo):
        self.total['archived'] += int(not undo)
        self.total['unarchived'] += int(undo)
        return True
    self.logger.info("Media id %s is not %s.", media_id, 'unarchived' if undo else 'archived')
    return False


def archive_medias(self, medias):
    broken_items = []
    if not medias:
        self.logger.info("Nothing to archive.")
        return broken_items
    self.logger.info("Going to archive %d medias." % (len(medias)))
    for media in tqdm(medias):
        if not self.archive(media):
            self.error_delay()
            broken_items = medias[medias.index(media):]
            break
    self.logger.info("DONE: Total archived %d medias." % self.total['archived'])
    return broken_items


def unarchive_medias(self, medias):
    broken_items = []
    if not medias:
        self.logger.info("Nothing to unarchive.")
        return broken_items
    self.logger.info("Going to unarchive %d medias." % (len(medias)))
    for media in tqdm(medias):
        if not self.unarchive(media):
            self.error_delay()
            broken_items = medias[medias.index(media):]
            break
    self.logger.info("DONE: Total unarchived %d medias." % self.total['unarchived'])
    return broken_items



import json
try:
    from unittest.mock import Mock, patch
except ImportError:
    from mock import Mock, patch

import requests

from instabot import Bot


class TestBot:
    def setup(self):
        self.USER_ID = 1234567
        self.USERNAME = 'test_username'
        self.PASSWORD = 'test_password'
        self.FULLNAME = 'test_full_name'
        self.TOKEN = 'abcdef123456'
        self.bot = Bot()
        self.prepare_api(self.bot)

    def prepare_api(self, bot):
        bot.api.is_logged_in = True
        bot.api.session = requests.Session()

        cookies = Mock()
        cookies.return_value = {
            'csrftoken': self.TOKEN,
            'ds_user_id': self.USER_ID
        }
        bot.api.session.cookies.get_dict = cookies
        bot.api.set_user(self.USERNAME, self.PASSWORD)


class TestBotAPI(TestBot):
    @patch('instabot.API.load_cookie')
    def test_login(self, load_cookie_mock):
        self.bot = Bot()

        load_cookie_mock.side_effect = Exception()

        def mockreturn(*args, **kwargs):
            r = Mock()
            r.status_code = 200
            r.text = '{"status": "ok"}'
            return r

        def mockreturn_login(*args, **kwargs):
            r = Mock()
            r.status_code = 200
            r.text = json.dumps({
                "logged_in_user": {
                    "pk": self.USER_ID,
                    "username": self.USERNAME,
                    "full_name": self.FULLNAME
                },
                "status": "ok"
            })
            return r

        with patch('requests.Session') as Session:
            instance = Session.return_value
            instance.get.return_value = mockreturn()
            instance.post.return_value = mockreturn_login()
            instance.cookies = requests.cookies.RequestsCookieJar()
            instance.cookies.update(
                {'csrftoken': self.TOKEN, 'ds_user_id': self.USER_ID})

            assert self.bot.api.login(username=self.USERNAME,
                                      password=self.PASSWORD)

        assert self.bot.api.username == self.USERNAME
        assert self.bot.user_id == self.USER_ID
        assert self.bot.api.is_logged_in
        assert self.bot.api.uuid
        assert self.bot.api.token

    def test_generate_uuid(self):
        from uuid import UUID
        generated_uuid = self.bot.api.generate_UUID(True)

        assert isinstance(UUID(generated_uuid), UUID)
        assert UUID(generated_uuid).hex == generated_uuid.replace('-', '')

    def test_set_user(self):
        test_username = "abcdef"
        test_password = "passwordabc"
        self.bot.api.set_user(test_username, test_password)

        assert self.bot.api.username == test_username
        assert self.bot.api.password == test_password
        assert hasattr(self.bot.api, "uuid")

    def test_reset_counters(self):
        keys = ['liked', 'unliked', 'followed', 'messages',
                'unfollowed', 'commented', 'blocked', 'unblocked']
        for key in keys:
            self.bot.total[key] = 1
            assert self.bot.total[key] == 1
        self.bot.reset_counters()
        for key in keys:
            assert self.bot.total[key] == 0
